#!/usr/bin/env python3
from __future__ import print_function, unicode_literals


"""traktorkjetting.py: stream/record in mp3/opus from traktor/mixxx"""
__version__   = "1.1"
__author__    = "ed <traktorkjetting@ocv.me>"
__url__       = "https://ocv.me/dev/?traktorkjetting.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019


r"""
this was thrown together in a few hours when i realized that traktor
still only supports streaming in vorbis (fairly low quality at that)
and somehow it worked on first try so i will never touch it again
...ok nevermind here's an update

it has been tested on MacOS Catalina 10.15,
also on Alpine Linux 3.10 with mixxx (and optional pulseaudio),
it should work on Windows and Linux with no changes

========================================================================
== if traktor ==========================================================

FIRST-TIME SETUP:
- traktor settings -> broadcasting:
    proxy = none
    address = 127.0.0.1
    port = 34108
    mount path = memes
    password = memes
    format = ogg vorbis, 44100 ghz, 64 kbit/s
- traktor settings -> mix recorder:
    source = internal (probably)
    split at size = 64 MB
    directory = c:\foo\bar
- optionally edit "Traktor Settings.tsi" in notepad and set this:
    <Entry Name="Recording.Splitsize" Type="1" Value="4"></Entry>
- create traktorkjetting-config.txt with the following contents:
    username:password
    icecast-address:port
    icecast-mountpoint
- so for example this (exactly three lines btw)
    source:hackme
    super.websight.com:8000
    main.mp3
- edit this script and set CONFIG_WAV_FOLDER = 'c:\foo\bar'

USAGE:
1) start traktor
2) start this script
3) in traktor, click the record button
4) in traktor, click the stream button

WHEN THIS SCRIPT CRASHES WHILE STREAMING:
1) restart this script
2) in traktor, doubleclick the stream bytton

though that should only happen if your machine is overloaded, like
if you're using a macbook air and browsing the internet while streaming

HOW IT WORKS:
traktor creates wav files which are consumed by this script,
traktor also creates a vorbis icecast stream which is sent to
this script (which pretends to be an icecast server)
just so it can collect the song tags and throw away the vorbis

========================================================================
== elif mixxx ==========================================================

FIRST-TIME SETUP:
- in mixxx preferences > recording,
    set output file format lossless wav pcm
	set output file format bit depth 16 bits
	set recordings directory c:\foo\bar
	set file splitting to the smallest possible value
	  (probably 60 minutes unless you patch mixxx)
- edit this script and set CONFIG_WAV_FOLDER = 'c:\foo\bar'
- edit this script and set TAGSOURCE = 'mixxx'

USAGE:
1) start mixxx and this script
2) in mixxx, click the record button

========================================================================
== end if ==============================================================

PS:
note that traktor/mixxx will write pcm audio to c:\foo\bar nonstop
so if you are worried about wearing out your SSD you should
consider making this a ramdisk

assuming this script doesn't crash it will delete each wav-file
as they've been sent to the server so you only need like 96 MB

if you use mixxx (and don't patch it) you'll get much larger files
so around 1300 MB will be used at any point in time

note that this script will delete every wav file in c:\foo\bar

PPS:
finally a short list of things i wish didn't have to be
which is currently just one thing (nice)

we use two separate instances of FFmpeg which are chained together,
one encodes to mp3, the other sends it to the destination icecast

first attempt was using the ffmpeg tee filter so one instance could
do all of the above however it's busted and won't let you set
icecast attributes so we have to snoop the mp3 data in the script
and bump the latency by several seconds in the process (nice)

PPPS:
a bunch of this code was stolen from https://github.com/9001/copyparty
just in case it looks familiar
"""


CONFIG_WAV_FOLDER = r'/Users/ed/Music/Traktor/Recordings'
TAGSOURCE = 'ice'

#CONFIG_WAV_FOLDER = r'/dev/shm/djwav'
#TAGSOURCE = 'mixxx'


import re
import os
import sys
import time
import socket
import base64
import struct
import threading
import http.client
import urllib.parse
import subprocess as sp
try:
    from queue import Queue
except:
    from Queue import Queue

from datetime import datetime, timedelta
import calendar


HTTPCODE = {
    200: "OK",
    206: "Partial Content",
    304: "Not Modified",
    400: "Bad Request",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    413: "Payload Too Large",
    422: "Unprocessable Entity",
    500: "Internal Server Error",
    501: "Not Implemented",
}


class Pebkac(Exception):
    def __init__(self, code, msg=None):
        super(Pebkac, self).__init__(msg or HTTPCODE[code])
        self.code = code


class Deets(object):
    def __init__(self, deetsfile):
        lines = []
        with open(deetsfile, 'rb') as f:
            for ln in f:
                lines.append(ln.decode('utf-8').rstrip())
        
        self.auth, self.ep, self.mount = lines
        self.host, self.port = self.ep.rsplit(':', 1)


class Tagpipe(object):
    def __init__(self):
        self.tags = Queue()
        #self.data = Queue(1024)


class TcpCli(object):
    def __init__(self, log, tagpipe, sck, addr):
        self.log = log
        self.tagpipe = tagpipe
        self.sck = sck
        self.addr = addr
        self.re_ctl = re.compile(br'[\x00-\x1f]')

        thr = threading.Thread(target=self.worker)
        thr.daemon = True
        thr.start()

    def terminate(self):
        try:
            self.sck.shutdown(socket.SHUT_RDWR)
            self.sck.close()
        except (OSError, socket.error) as ex:
            if ex.errno not in [107, 9]:
                # 107 Transport endpoint not connected
                #   9 Bad file descriptor
                raise
        
    def reply(self, body, status=200, mime="text/html", headers=[]):
        response = [
            "HTTP/1.1 {} {}".format(status, HTTPCODE[status]),
            "Content-Type: " + mime,
            "Content-Length: " + str(len(body)),
            "Connection: Close",
        ]

        response.extend(headers)
        response_str = "\r\n".join(response).encode("utf-8")
        try:
            self.s.sendall(response_str + b"\r\n\r\n" + body)
        except:
            raise Pebkac(400, "client disconnected before http response")

        return body

    def loud_reply(self, body, *args, **kwargs):
        self.log(body.rstrip())
        self.reply(b"<pre>" + body.encode("utf-8"), *list(args), **kwargs)
        self.terminate()

    def read_header(self):
        ret = b""
        while True:
            if ret.endswith(b"\r\n\r\n"):
                break
            elif ret.endswith(b"\r\n\r"):
                n = 1
            elif ret.endswith(b"\r\n"):
                n = 2
            elif ret.endswith(b"\r"):
                n = 3
            else:
                n = 4

            buf = self.sck.recv(n)
            if not buf:
                if not ret:
                    return None

                raise Pebkac(
                    400,
                    "protocol error while reading headers:\n"
                    + ret.decode("utf-8", "replace"),
                )

            ret += buf

            if len(ret) > 1024 * 64:
                raise Pebkac(400, "header 2big")

        return ret[:-4].decode("utf-8", "replace").split("\r\n")
    
    r"""
    4f 67 67 53  00 00 00 00   00 00 00 00  00 00 5c ad   OggS ....  .... ..\.
    2b 37 01 00  00 00 6e 01   bb 1d[16]a9  ff ff ff ff   +7.. ..n.  »..© ÿÿÿÿ
    ff ff ff ff  ff ff ff ff   ff ff ff ff  ff ff ff ff   ÿÿÿÿ ÿÿÿÿ  ÿÿÿÿ ÿÿÿÿ
    51 03 76 6f  72 62 69 73   35 00 00 00  58 69 70 68   Q.vo rbis  5... Xiph
    2e 4f 72 67  20 6c 69 62   56 6f 72 62  69 73 20 49   .Org  lib  Vorb is I
    20 32 30 31  38 30 33 31   36 20 28 4e  6f 77 20 31    201 8031  6 (N ow 1
    30 30 25 20  66 65 77 65   72 20 73 68  65 6c 6c 73   00%  fewe  r sh ells
    29 03 00 00  00 28 00 00   00 45 4e 43  4f 44 45 52   )... .(..  .ENC ODER
    3d 4e 61 74  69 76 65 20   49 6e 73 74  72 75 6d 65   =Nat ive   Inst rume
    6e 74 73 20  4d 65 64 69   61 20 4c 69  62 72 61 72   nts  Medi  a Li brar
    79 14 00 00  00 41 52 54   49 53 54 3d  44 44 52 4b   y... .ART  IST= DDRK
    69 72 62 79  28 49 53 51   29 1c 00 00  00 54 49 54   irby (ISQ  )... .TIT
    4c 45 3d 54  65 61 63 68   20 4d 65 20  5b 48 65 72   LE=T each   Me  [Her
    65 27 73 20  48 6f 77 21   5d 01 05 76  6f 72 62 69   e's  How!  ]..v orbi
    73 46 42 43  56 01 00 08   00 00 00 31  4c 20 c5 80   sFBC V...  ...1 L Å.
    d0 90 55 00  00 10 00 00   60 24 29 0e  93 66 49 29   Ð.U. ....  `$). .fI)
    a5 94 a1 28  79 98 94 48   49 29 a5 94  c5 30 89 98   ¥.¡( y..H  I)¥. Å0..
    94 89 c5 18  63 8c 31 c6   18 63 8c 31  c6 18 63 8c   ..Å. c.1Æ  .c.1 Æ.c.
    """

    def get_text(self, txt):
        m = self.re_ctl.search(txt)
        if m:
            return txt[:m.start()]
        else:
            return txt

    def worker(self):
        try:
            headers = self.read_header()
        except Pebkac as ex:
            return self.loud_reply(str(ex), status=ex.code)
        
        self.log('ice-in', 'Traktor connected')
        self.sck.sendall(b'HTTP/1.0 200 OK\r\n\r\n')
        prev = b''
        while True:
            inc = self.sck.recv(4096)
            if not inc:
                self.log('ice-in', 'Traktor disconnected')
                break
            
            #if not self.tagpipe.data.full():
            #    self.tagpipe.data.put(inc)
            
            buf = prev + inc
            prev_len = len(prev)
            prev = inc

            oa = buf.find(b'\x00ARTIST=')
            ot = buf.find(b'\x00TITLE=')
            #print(ofs1, ofs2)
            
            if oa < 0 or ot < 0:
                continue
            
            if oa >= prev_len and ot >= prev_len:
                continue
            
            la = struct.unpack('<L', buf[oa-3:oa+1])[0]
            lt = struct.unpack('<L', buf[ot-3:ot+1])[0]
            #self.log('ice-in', '{}, {}'.format(la, lt))
            artist = buf[oa+8:oa+la+1].decode('utf-8', 'replace')
            title = buf[ot+7:ot+lt+1].decode('utf-8', 'replace')

            self.tagpipe.tags.put([artist, title])
            #self.log('ice-in', 'artist: {}'.format(artist))
            #self.log('ice-in', 'title: {}'.format(title))


class TcpSrv(threading.Thread):
    def __init__(self, log, tagpipe):
        self.log = log
        self.tagpipe = tagpipe

        self.ip = '127.0.0.1'
        self.port = 34108

        self.srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.srv.bind((self.ip, self.port))
        except (OSError, socket.error) as ex:
            if ex.errno == 98:
                raise Exception(
                    "\033[1;31mport {} is busy on interface {}\033[0m".format(
                        self.port, self.ip
                    )
                )
            raise
        
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        self.srv.listen(8)
        self.log("tcpsrv", "listening @ {0}:{1}".format(self.ip, self.port))

        while True:
            #self.log("tcpsrv", "-" * 2 + "C-acc1")
            sck, addr = self.srv.accept()
            #self.log(str(addr), "-" * 3 + "C-acc2")
            TcpCli(self.log, self.tagpipe, sck, addr)


class MixxxTags(threading.Thread):
    def __init__(self, log, tagpipe):
        self.log = log
        self.tagpipe = tagpipe
        
        threading.Thread.__init__(self)
        self.daemon = True
    
    def run(self):
        last_tag = None
        while True:
            time.sleep(2)
            tag = self.gettags()
                
            if tag is None:
                self.log('mixxx-tags', 'nulltag')
                continue
            
            if last_tag == tag:
                continue
            
            last_tag = tag
            
            try:
                tag, _ = tag.split(' | Mixxx ', 1)
            except:
                pass
            
            try:
                a, b = tag.split(' `` ')
                tag = [a, b]
            except:
                try:
                    a, b = tag.split(' - ')
                    tag = [a, b]
                except:
                    tag = ['', tag]
            
            self.tagpipe.tags.put(tag)
    
    def gettags(self):
        try:
            cmd = ['xdotool', 'search', '-name',
                r' \| Mixxx [0-9][0-9\.]+[0-9] x64$',
                'getwindowname']
            
            return sp.check_output(cmd, stderr=sp.DEVNULL).decode('utf-8', 'replace')
        except:
            return None

    
class Tagpusher(threading.Thread):
    def __init__(self, log, tagpipe, deets, t0):
        self.log = log
        self.tagpipe = tagpipe
        self.deets = deets
        self.t0 = t0

        threading.Thread.__init__(self)
        self.daemon = True
    
    def run(self):
        tagfn = 'rec-{}.tags'.format(self.t0)
        auth = "Basic " + base64.b64encode(self.deets.auth.encode('utf-8')).decode('utf-8')
        while True:
            tag = self.tagpipe.tags.get()
            cue = [str(time.time())] + tag
            with open(tagfn, 'wb') as f:
                f.write('\n'.join(cue).encode('utf-8', 'replace') + b'\n\n')
            
            url = "https://{ep}/admin/metadata?mode=updinfo&mount=/{mount}&charset=UTF-8&song={tag}".format(
                ep = self.deets.ep,
                mount = self.deets.mount,
                tag = urllib.parse.quote_plus(' - '.join(tag).encode('utf-8')))

            conn = http.client.HTTPConnection(self.deets.host, self.deets.port, timeout=3)
            conn.request('GET', url, headers={
                "Authorization": auth
            })
            r = conn.getresponse()
            self.log('ice-tag', 'server response: {}'.format(r.status))
            
            rt = r.read().decode('utf-8')
            rt = rt.replace('<?xml version="1.0"?>', '')
            rt = rt.replace('\r', '').replace('\n', '')
            self.log('ice-tag', rt)


class Wavreader(threading.Thread):
    def __init__(self, log, wavdir):
        self.log = log
        self.wavdir = wavdir
        self.q = Queue(128)

        threading.Thread.__init__(self)
        self.daemon = True
    
    def getwavs(self):
        ret = []
        for fn in os.listdir(self.wavdir):
            if fn.endswith('.wav'):
                path = os.path.join(self.wavdir, fn)
                inf = os.stat(path)
                if inf.st_size > 44:
                    ret.append([inf.st_mtime, inf.st_size, path])
        
        ret.sort(reverse=True)
        return ret

    def run(self):
        self.log('wavreader', 'hello')
        warn_about_empty_input_folder = True
        last_fn = None
        last_pos = None
        first_open = True
        prev_path = None
        prev_depart_at = None
        while True:
            files = self.getwavs()
            if not files:
                if warn_about_empty_input_folder:
                    warn_about_empty_input_folder = False
                    self.log('wavreader', 'no wav files in {}'.format(self.wavdir))
                
                time.sleep(0.3)
                continue
            
            warn_about_empty_input_folder = True

            resume = False
            for _, sz, fn in files:
                if fn == last_fn and sz > last_pos:
                    resume = True
                    break
            
            if not resume:
                for _, sz, fn in files[2:]:
                    self.log('wavreader', 'deleting {}'.format(fn))
                    os.remove(fn)
                
                _, sz, fn = files[0]
            
            if fn != last_fn:
                self.log('wavreader', 'open {}'.format(fn))
                prev_depart_at = time.time()
                prev_path = last_fn

            if prev_path and time.time() - prev_depart_at > 10:
                self.log('wavreader', 'deleting {}'.format(prev_path))
                os.remove(prev_path)
                prev_path = None

            # header: 44
            # 44100hz, 16bit, 2ch
            # 44100 * 2 * 2
            
            if fn == last_fn:
                ofs = last_pos
                if sz > last_pos:
                    time.sleep(0.05)
                else:
                    time.sleep(0.1)
            else:
                last_fn = fn
                ofs = max(44, int((sz - 44) / 4) * 4 - int(44100 * 2 * 2 * 0.5))
                
                bufsz_seek = ofs / (44100. * 2 * 2)
                bufsz_start = sz / (44100. * 2 * 2)
                msg = 'SKIPPING!'
                
                # do not skip start of file after a split
                # unless we are more than 3sec late
                if not first_open \
                and ofs < 44100 * 2 * 2 * 3:
                    msg = 'playing from start'
                    ofs = 44
                
                self.log('wavreader', 'bufsz {:.2f} / {:.2f}, {}'.format(
                    bufsz_start, bufsz_seek, msg))
                
                first_open = False
            
            #self.log('wavreader', 'seek {} ({} remains)'.format(ofs, sz - ofs))

            with open(fn, 'rb', 512*1024) as f:
                f.seek(ofs)
                while True:
                    last_pos = f.tell()
                    buf = f.read(4096)
                    if not buf:
                        break
                    
                    #self.log('wavreader', 'read {}'.format(len(buf)))
                    self.q.put(buf)
            
            #self.log('wavreader', 'eof')


class FFmpeg(threading.Thread):
    def __init__(self, log, deets, waver, t0):
        self.log = log
        self.deets = deets
        self.waver = waver
        self.t0 = t0

        self.ep = '{}@{}/{}'.format(
            self.deets.auth,
            self.deets.ep,
            self.deets.mount
        )

        threading.Thread.__init__(self)
        self.daemon = True
    
    def run(self):
        url = 'icecast://{}'.format(self.ep)
        tee_opts = ':'.join([
            r'content_type=audio\/mpeg',
            r'ice_name=memes',
            r'legacy_icecast=1'
        ])
        
        cmd1 = [
            'ffmpeg',
            '-re',
            '-v', 'warning',
            '-hide_banner',
            '-ac', '2',
            '-f', 's16le',
            '-ar', '44100',
            '-i', '-',
            '-acodec', 'libmp3lame',
            '-b:a', '192k',
            '-f', 'mp3', '-'
        ]
        cmd2 = [
            'ffmpeg',
            '-hide_banner',
            '-i', '-',
            '-codec', 'copy',
            '-content_type', 'audio/mpeg',
            '-ice_name', 'memes',
            '-legacy_icecast', '1', url
        ]
        self.log('ffmpeg', '\033[44m \033[0m'.join(cmd1))
        self.log('ffmpeg', '\033[44m \033[0m'.join(cmd2
            ).replace(self.deets.ep, "<SECRET>"
            ).replace(self.deets.auth, "<SECRET>"))
        
        p1 = sp.Popen(cmd1, stdin=sp.PIPE, stdout=sp.PIPE)
        p2 = sp.Popen(cmd2, stdin=sp.PIPE)
        
        thr = threading.Thread(target=self.badmeme, args=(p1.stdout, p2.stdin, ))
        thr.daemon = True
        thr.start()

        while True:
            if p1.poll() is not None:
                raise Exception('\n\nFFMPEG ENC DIED\n')
            
            if p2.poll() is not None:
                raise Exception('\n\nFFMPEG ICE DIED\n')
            
            #self.log('ffmpeg', 'get')
            buf = self.waver.q.get()
            p1.stdin.write(buf)

    def badmeme(self, pin, pout):
        with open('cap-{}.mp3'.format(self.t0), 'wb') as f:
            while True:
                buf = pin.read(4096)
                pout.write(buf)
                f.write(buf)


class Logger(object):
    def __init__(self):
        self.log_mutex = threading.Lock()
        self.next_day = 0

    def logfunc(self, src, msg):
        with self.log_mutex:
            now = time.time()
            if now >= self.next_day:
                dt = datetime.utcfromtimestamp(now)
                print("\033[36m{}\033[0m".format(dt.strftime("%Y-%m-%d")))

                # unix timestamp of next 00:00:00 (leap-seconds safe)
                day_now = dt.day
                while dt.day == day_now:
                    dt += timedelta(hours=12)

                dt = dt.replace(hour=0, minute=0, second=0)
                self.next_day = calendar.timegm(dt.utctimetuple())

            ts = datetime.utcfromtimestamp(now).strftime("%H:%M:%S.%f")[:-3]

            fmt = "\033[36m{} \033[33m{:21} \033[0m{}\033[J"
            msg = fmt.format(ts, src, msg)
            try:
                print(msg)
            except UnicodeEncodeError:
                print(msg.encode("utf-8", "replace").decode())


def main():
    t0 = time.time()
    log = Logger().logfunc
    log('core', 'will stream wav files from {}'.format(CONFIG_WAV_FOLDER))

    deetsfile = 'traktorkjetting-config.txt'
    if len(sys.argv) > 1:
        deetsfile = sys.argv[1]
    
    log('core', 'using icecast config in {}'.format(deetsfile))
    deets = Deets(deetsfile)

    tagpipe = Tagpipe()
    
    if TAGSOURCE == 'mixxx':
        MixxxTags(log, tagpipe).start()
    elif TAGSOURCE == 'ice':
        TcpSrv(log, tagpipe).start()
    else:
        raise Exception('bad TAGSOURCE, change it and try again')

    tagpusher = Tagpusher(log, tagpipe, deets, t0)
    tagpusher.start()

    waver = Wavreader(log, CONFIG_WAV_FOLDER)
    waver.start()

    ffmpeg = FFmpeg(log, deets, waver, t0)
    ffmpeg.start()

    while True:
        time.sleep(60)


if __name__ == '__main__':
    main()
